/*
   getpvalue.c -- subroutine to compute gamma and p from a histogram,
                  taking the middle 80% of values.
*/
/*
   Author: Paul Hardy, U.S.C., December 1995.

   Version 1.0 release date: December 1995.
*/
#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include "seqaln.h"


getpvalue(seq1,seq2,scoring,results,io)
struct SEQALN_SEQUENCE *seq1,*seq2;
struct SEQALN_CONSTANTS *scoring;
struct SEQALN_RESULTS *results;
struct SEQALN_IO *io;
{
int low, high;			/* indices into histogram array           */
int i, k;			/* indices for histogram and x,y, resp.   */
int nscores, tmpscores;		/* number and sum of scores used          */
int tenth;			/* one-tenth of the scores, to chop ends  */
int subtotal;			/* subtotal of histogram scores seen      */
int xn;				/* number of unique scores used           */
int nalloc;			/* number of unique scores in x & y       */
double *x, *y;			/* x, y vectors for lin. regression       */
double histpt;			/* one element of the histogram           */
double logn;			/* log of (scores used + 1) or (nscores)  */
double slope;			/* log of the p-value                     */
double intercept;		/* log(gamma * seq1->len * seq2->len)     */
double coeff;			/* reliability coefficient                */
double log(),exp(),pow(),infinity();

/*
   Determine total number of scores: # simulations * # scores from each.
*/
nscores = results->pvalue->nsims * (results->pvalue->ndeclumps + 1);
/*
   Get lowest score found.
*/
for (low = 0;
     low < results->pvalue->nhist && results->pvalue->hist[low] == 0;
     low++);
/*
   Get highest score found and number of unique scores at same time.
*/
xn = 0;
tmpscores = 0;
for (high=low; tmpscores < nscores && high < results->pvalue->nhist; high++) {
   if (results->pvalue->hist[high] > 0) {
      xn++;			/* one more unique score found */
      tmpscores += results->pvalue->hist[high];	/* add to total scores found */
    }
 }
high--;
nalloc = xn;
/*
   toss out lower 10% & upper 10% of data
*/
if (scoring->allscores == 0) {
   tenth = nscores * 0.1;
   /* Adjust low and high, and the number of scores in each, accordingly */
   tmpscores = tenth;
   while (tmpscores > 0) {
      nalloc--;			/* using one less unique score */
      low++;
      tmpscores -= results->pvalue->hist[low];
    }
   while (results->pvalue->hist[low] == 0) /* get next non-zero entry */
      low++;

   for (i=0, tmpscores = 0; tmpscores < tenth; i++) {
      tmpscores += results->pvalue->hist[i];
      if (tmpscores >= tenth) {	/* lower 10% can include this point */
	 low = i+1;
       }
    }
   while (results->pvalue->hist[low] == 0 && low < results->pvalue->nhist)
      low++;
   /* Adjust low and high, and the number of scores in each, accordingly */
   for (i=results->pvalue->nhist, tmpscores = 0; tmpscores < tenth; i--) {
      tmpscores += results->pvalue->hist[i];
      if (tmpscores >= tenth) {
	 high = i-1;
       }
    }
   while (results->pvalue->hist[high] == 0 && high > low)
      high--;
 }

xn = high - low + 1;
if ((x = (double *)malloc(xn * sizeof(double))) == (double *)NULL) {
   (void)fprintf(io->STDERR,
           "Error: can't allocate array for linear regression.\n\n");
   exit(-1);
 }
if ((y = (double *)malloc((xn+1) * sizeof(double))) == (double *)NULL) {
   (void)fprintf(io->STDERR,
             "Error: can't allocate array for linear regression.\n\n");
   exit(-1);
 }

k = 0;				/* counter in x,y arrays */
tmpscores = 0;
/* if using all scores, avoid taking log(0) */
if (scoring->allscores || high == results->pvalue->nhist) {
   nscores++;
 }
logn = log((double)nscores);

for (i=0; i<=high; i++) {
   tmpscores += results->pvalue->hist[i];
   if (i >= low && results->pvalue->hist[i] > 0) {
      histpt = tmpscores;
      histpt = log(histpt);
      /* The below condition should always be true; check just in case */
      if (logn > histpt) {
	 x[k] = i;
	 y[k] = log(logn - histpt);
	 if (scoring->Debug) {
	    (void)fprintf(io->STDOUT,
                   "%6d  %6g  %6d %6g\n",k,x[k],tmpscores,y[k]);
	    }
	 k++;
       }
    }
 }
xn = k;
/* Compute the linear regression values. */
linreg(xn,x,y,&slope,&intercept,&coeff);

results->pvalue->gamma  = exp(intercept) /
  ((double)seq1->len * (double)seq2->len);
results->pvalue->p = exp(slope);
results->pvalue->coeff = coeff; /* params bad if between -1 & 1 */

if (scoring->rptptable) {
  (void)fprintf(io->STDOUT,
            "\nTable of observed and expected P(t <= score)\n\n");
  (void)fprintf(io->STDOUT,"            Prob.     Prob.\n");
  (void)fprintf(io->STDOUT,"   Score  Observed  Expected\n");
  for (i=0; i<xn; i++) {
    (void)fprintf(io->STDOUT,"%7g   %8.6f  %8.6f\n",
	   x[i],
	   exp(-exp(y[i])),
	   exp(-results->pvalue->gamma * seq1->len * seq2->len *
	       pow(results->pvalue->p,x[i])));
  }
  (void)fprintf(io->STDOUT,"\n");
}
return;
}
