/* function to load the next sequence in file name into sequence struct;
   return value is length of sequence.  Opens the file the first time called.
*/
/*
   Author: Paul Hardy, U.S.C., December 1995.

   Version 1.17 (PH): Now supports ONLY the FASTA format
                      Can read an entire database of sequences
                      First argument, "fn", moved into sequence->fname
   Version 1.16 (PH): Added to FASTA format to ignore lines starting with ';'.
   Version 1.1 (PH): February 1996: if first non-blank character is '>',
                     assume entire file is in FASTA format.
   Version 1.0 release date: December 1995.
*/
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include "seqaln.h"

int load_seq(sequence, scoring, io)
     struct SEQALN_SEQUENCE *sequence;
     struct SEQALN_CONSTANTS *scoring;
     struct SEQALN_IO *io;
{
FILE *f;
int fasta_format;
int c;
char *a,*b;
int i,j;
int seqalloc;    /* Amount of memory to allocate for [sub]sequence */
char *newmem();
long ftell();

printf("Entered load_seq for %s\n",sequence->fname);
	if (sequence->fname == (char *)NULL) {
	   (void)fprintf(io->STDERR, "load_seq: sequence file name not set.\n");
	   return(-1);
	   }
        if (strcmp(sequence->fname,"-") == 0) {  /* "-" denotes stdin */
           sequence->fptr = io->STDIN;
           sequence->fstart = 0;
           sequence->fend = 0;
           }
        else if (sequence->fptr == (FILE *)NULL) {  /* Initial file access */
           sequence->fstart = 0;
           sequence->fend = 0;
           /* Look for '@' at end of file name, which signifies seq range */
           for (i=strlen(sequence->fname) - 1;
                i > 0 && sequence->fname[i] != '@';
                i--);
           if (sequence->fname[i] == '@') {  /* Read possible start/stop ORF pair */
	      sequence->fname[i] = '\0';     /* Terminate file name here          */
              for (++i; sequence->fname[i] && !isdigit(sequence->fname[i]); i++);
              /* Look for end of starting location */
              for (j=i; j<strlen(sequence->fname) && isdigit(sequence->fname[j]); j++);
	      if ((j+1) < strlen(sequence->fname)) {
                 sequence->fname[++j] = '\0';     /* Terminate start loc here     */
                 }
   printf("Starting loc is %d to %d\n", i, j);
              /* Try to allow "reasonable" starting location */
              sequence->fstart = atoi(&sequence->fname[i]);
              if (sequence->fstart < 0)
                 sequence->fstart = 0;
   printf("Starting at %d\n", sequence->fstart);
              /* Try to find an ending location specified */
              for (++j;
                   j < strlen(sequence->fname) && !isdigit(sequence->fname[j]);
                   j++);
	      if (j < strlen(sequence->fname)) {  /* Found an ending location */
	         sequence->fend = max(sequence->fstart, atoi(&sequence->fname[j]));
	         }
	      }
           }
	if (sequence->fptr == (FILE *)NULL) {
	   f=fopen(sequence->fname, "r");
	   if (f==NULL) {
	      (void)fprintf(io->STDERR, "bad sequence file %s\n", sequence->fname);
	      sequence->len = 0;
	      return(-1);
	      }
      printf("Opened sequence file %s successfully.\n",sequence->fname);
	   sequence->fptr = f;
	   sequence->fposn = 0L;  /* start of file */
           }
	else {
	   f = sequence->fptr;
	   }
	/* Read from start of sequence */
	(void)fseek(f, sequence->fposn, SEEK_SET);
        i=0;
        while ((c = fgetc(f)) != '>' && c != EOF);
	while ((c = fgetc(f)) == '>'); /* can begin w/many >'s */
        /* Read the sequence name into the name array */
	for (i=0; i<MAXSEQNAME-1 && c != EOF && c != '\n'; i++) {
           if (c != '\r')
              sequence->name[i] = c;
           else
              sequence->name[i] = '\0';
           c = fgetc(f);
	   }
	sequence->name[i] = '\0';  /* nullify the newline */
	while ((c=fgetc(f)) == ';') { /* FASTA annotation line */
	   while ((c=fgetc(f)) != '\n'); /* read to end of line */
	   }
        if (c != '\r')
	   ungetc(c,f);  /* part of seq -- put back to read below */

        /* Determine length of sequence. */
	sequence->fposn = ftell(f);  /* We'll return after memory alloc */
	for (j=0; (c=fgetc(f)) != EOF && c != '>' ; ) {
	   if (!isspace(c))
	      j++;
	   }
	sequence->fend = min(sequence->fend, j);  /* Don't overrun seq */
	sequence->len = seqalloc  /* Memory required */
	              = sequence->fstart - sequence->fend + 1;
        /* Allocate memory to hold sequence if not enough alloc'd. */
	if (sequence->len > sequence->maxlen) {
	   sequence->resized = 1; /* mark size change, for results_init */
	   if (sequence->seq != (char *)NULL)
	      free(sequence->seq);
	   if (sequence->code != (char *)NULL)
	      free(sequence->code);
	   sequence->seq = newmem(sequence->len+2, sizeof(char));
	   sequence->code = newmem(sequence->len+2, sizeof(char));
	   sequence->maxlen = sequence->len;  /* Increased alloc'd seq memory */
           }
        else {
           sequence->resized = 0;
           }
	a = sequence->seq;
	b = sequence->code;
	/* Position to start of sequence string */
	(void)fseek(f, sequence->fposn, SEEK_SET);
        /* Read sequence into allocated space. */
	j = 0; /* Absolute position in sequence */
	for (a++, b++; (c=fgetc(f)) != EOF && c != '>' && c != ';' ; ) {
	   if (c && !isspace(c)) {
	      j++;
	      if (j >= sequence->fstart &&
	          j <= sequence->fend) {
	         *a = toupper(c);   /* Sequence     */
                 *b = asc2num(*a);  /* Numeric code */
                 a++;
                 b++;
	         }
	      }
	   }
	*a = '\0';
	while (c != EOF && c != '>') {
	   c = fgetc(f);
	   }
	if (c == '>') {
	   ungetc(c,f);
	   }
	sequence->fposn = ftell(f);  /* Start of new sequence */
        if (sequence->len <= 0) {
	   if (c != EOF) {
              fprintf(io->STDERR,"Null sequence encountered.\n\n");
	      }
           sequence->len = 0;
           sequence->seq = "";
           sequence->code = "";
           }
	if (j == 0 && c == EOF) /* If no more data remains to be read */
	   sequence->len = -1;

 printf("Sequence start, end (%d, %d): %s\n",
        sequence->fstart, sequence->fend, &sequence->seq[1]);
	return(sequence->len);  /* Length of sequence read; < 0 = EOF */
}
