/* ptrace.c -- construct a profile traceback list */
/*
   Author: Paul Hardy, U.S.C., 1995, 1997.

   Version 2.0 (transposed profile matrix): March 1997.
   Version 1.0 release date: December 1995.
*/
#include <stdio.h>
#include "seqaln.h"

/* Macros for use in the plot() subroutine */

#define append_score(X,Y)  (scorel = \
			    (SEQALN_TRACE *)newmem(1, sizeof(SEQALN_TRACE)), \
                            scorel->i = i,                               \
			    scorel->j = j,                               \
                            scorel->next = results->trace,               \
			    results->trace = scorel)

/*
   Check the match/mismatch (diagonal) direction.
*/
#define scandiag()  if ((!results->aligned || !results->aligned[i][j]) &&     \
                        i>0 && j>0 && results->S[i][j]==results->S[i-1][j-1]+ \
	                scoring->deltafn(j,seq->code[i],pro)) {               \
		       results->align_len++; i--; j--; foundalign=1; }

/*
   Scan up from current position.
*/
#define scanup()    if (i > 0) {                                              \
		       tmplast = results->S[i][j];                            \
		       if (alpha_eq_beta) { /* just check one position up   */\
			  if (results->S[i-1][j]+scoring->alpha == tmplast) { \
                            results->align_len++;                             \
			    i--;                                              \
			    foundalign = 1;                                   \
			  }                                                   \
			}                                                     \
		       else {	/* check to start of column if necessary */   \
			 nexti = i;                                           \
			 tmpgap = scoring->alpha - scoring->beta;             \
			 Slast = results->S[i][j];                            \
			 do {                                                 \
			   tmpgap += scoring->beta;                           \
			   tmplast = results->S[--nexti][j];                  \
			 } while (nexti>0 && Slast != tmplast + tmpgap);      \
			 if (Slast == tmplast + tmpgap) {                     \
                           results->align_len += i - nexti;                   \
			   i = nexti;                                         \
			   foundalign = 1;	/* found a deletion */        \
			 }                                                    \
		       }                                                      \
		     }

/*
   Scan left from current position.
*/
#define scanleft()  if (j > 0) {                                              \
		       tmplast = results->S[i][j];                            \
		       if (palpha_eq_pbeta) { /* just check one position up */\
			  if (results->S[i][j-1]+scoring->palpha == tmplast) {\
                            results->align_len++;                             \
			    j--;                                              \
			    foundalign = 1;                                   \
			  }                                                   \
			}                                                     \
		       else {	/* check to start of column if necessary */   \
			 nextj = j;                                           \
			 tmpgap = scoring->palpha - scoring->pbeta;           \
			 Slast = results->S[i][j];                            \
			 do {                                                 \
			   tmpgap += scoring->pbeta;                          \
			   tmplast = results->S[i][--nextj];                  \
			 } while (nextj>0 && Slast != tmplast + tmpgap);      \
			 if (Slast == tmplast + tmpgap) {                     \
                           results->align_len += j - nextj;                   \
			   j = nextj;                                         \
			   foundalign = 1;	/* found a deletion */        \
			 }                                                    \
		       }                                                      \
		     }

/*
   Construct a traceback.

   results->bestS, results->bestSi and results->bestSj must be set before
   this routine is called, and of course the scoring matrix, results->S,
   must have been computed.
*/

void ptrace(seq, pro, scoring, results, io)
	struct SEQALN_SEQUENCE *seq;
        struct SEQALN_PROFILE *pro;
	struct SEQALN_CONSTANTS *scoring;
	struct SEQALN_RESULTS   *results;
        struct SEQALN_IO *io;
{
int i,j;                /* Position variables for seq & pro resp.    */
int alpha_eq_beta;	/* If scoring->alpha == scoring->beta        */
int palpha_eq_pbeta;	/* If pro->alpha == pro->beta                */
int Slast;		/* Score of previous cell in traceback       */
int tmplast;		/* For scoring multiple character deletions  */
int tmpgap;		/* For accumulated gap score for indel check */
int nexti, nextj;  	/* For checking multiple character deletions */
int foundalign;		/* For checking multiple character deletions */
SEQALN_TRACE *scorel;	/* Holding variable for next element in list */

results->trace = (SEQALN_TRACE *)NULL;
results->align_len = 0;
i = results->bestSi;
j = results->bestSj;
/*
   Set alpha_eq_beta and palpha_eq_pbeta for efficient evaluation.
*/
alpha_eq_beta  = scoring->alpha == scoring->beta;
palpha_eq_pbeta = pro->alpha == pro->beta;

while ((scoring->ptype == LOCAL  && results->S[i][j] > 0) ||
       (scoring->ptype == FIT    && i > 0)                ||
       (scoring->ptype == OVER && i > 0 && j > 0) ||
       (scoring->ptype == GLOBAL && (i > 0 || j > 0))) {

  append_score(i,j);
  if (scoring->Debug) {
    (void)fprintf(io->STDOUT,"(%d,%d)==>\n",i,j);
  }
  foundalign = 0;	/* Haven't aligned this letter pair yet */

  /*
    First scan direction.
  */
  if (scoring->envelope) {	/* tracing upper/lower alignment envelope */
    if (scoring->traceupper) {
      scanup();
    }
    else {
      scanleft();
    }
  }
  else {
    scandiag();			/* checking match/mismatch first */
  }

  /*
    Second scan direction.
  */
  if (!foundalign) {
    if (scoring->envelope) {	/* checking match/mismatch second */
      scandiag();
    }
    else {
      if (scoring->traceupper) {
	scanup();
      }
      else {
	scanleft();
      }
    }

    /*
      Third scan direction -- the indel direction we don't prefer.
    */
    if (!foundalign) {
      if (scoring->traceupper) {
	scanleft();
      }
      else {
	scanup();
      }
    }				/* end scan in third direction  */
  }				/* end scan in second direction */

  /*
    Should have found something by now -- checked all three directions.
  */
  if (!foundalign) { /* checked all directions & didn't get anywhere--abort */
    (void)fprintf(io->STDERR,
                "Traceback routine failed at %d %d.\n", i, j);
    (void)fprintf(io->STDERR,
                 "   score is %d; probably should be max(%d+%d*%d,%d,%d).\n",
                        results->S[i][j],
                        results->S[i-1][j-1],
                        scoring->deltafn(j, seq->code[i], scoring),
                        results->aligned ? !results->aligned[i][j] : 1,
                        results->S[i-1][j] + scoring->palpha,
                        results->S[i][j-1] + scoring->palpha);
    (void)fprintf(io->STDERR,
                  "Began traceback with score of %d at %d %d.\n",
	    results->bestS, results->bestSi, results->bestSj);
    print_trace(scoring,results,io);
    i = j = 0;		/* so loop will end, even if LOCAL since S[0][0]==0 */
    results->trace = (SEQALN_TRACE *)NULL;
  }
}	    /* end while */

if (scoring->Debug) {
  (void)fprintf(io->STDOUT,"(%d,%d).\n",i,j);
}

/*
   We encountered impossible values for i and/or j.
*/
if (i < 0 || j < 0) {
  (void)fprintf(io->STDERR,
            "Traceback parameters out of range at %d %d.\n",i,j);
  (void)fprintf(io->STDERR, "Began traceback at %d %d:\n",
	  results->bestSi, results->bestSj);
  print_trace(scoring,results,io);
  results->trace = (SEQALN_TRACE *)NULL;
}

return;
}
