/*
   results_init() -- function to allocate and initialize matrix memory,
                     depending upon the type of sequence alignment to be
                     performed.

                     What matters:
		       Determines matrix size:
			 scoring->btype = { BAND, NONBAND, TANDEM, SELF }
		       Determines boundary initialization:
		         scoring->ptype = { GLOBAL, FIT, LOCAL, OVER }
*/
/*
   Author: Paul Hardy, U.S.C., December 1995.

   Version 2.0, February 1997 (Paul Hardy):
                Now reallocates/re-initializes matrices
                for reading multiple sequences from databases.
   Version 1.0 release date: December 1995.
*/
#include <stdio.h>
#include <malloc.h>
#include "seqaln.h"

results_init(seq1,seq2,scoring,results,io)
  struct SEQALN_SEQUENCE *seq1,*seq2; /* 1st & 2nd sequences to be compared */
  struct SEQALN_CONSTANTS *scoring; /*Parameters related to scoring operation*/
  struct SEQALN_RESULTS *results;   /* Computation results. */
  struct SEQALN_IO *io;

  {
  int i,j,tscore;
  int Sresized;			/* Denotes whether or not S changed size */
  int *p;
  char *dtmp;
  int retval;
  int infty;
  int collen;                   /* for differentiating LINEAR, SQUARE     */
  int rowlen;			/* for differentiating band, normal, etc. */
  int gap();
  char *malloc(), *calloc();

  /*
     Copy from scoring structure; re-initialized for each new sequence.
  */
  results->naligns = scoring->naligns;
  /*
     Determine the second dimension of the scoring matrix.
  */
  collen = scoring->memuse == LINEAR ? 2 : seq1->len + 1;
  rowlen = seq2->len;
  if (scoring->btype != TANDEM) {
     rowlen++;
  }
  Sresized = (seq1->resized && scoring->memuse != LINEAR) || seq2->resized;

  /*
     Initialize scoring matrix.
  */
  if (Sresized) {
    if (results->S != (int **)NULL) {
      if (results->S[0] != (int *)NULL) {
	free(results->S[0]);
      }
      free(results->S);
    }
    if ((results->S = (int **)malloc(collen * (sizeof(int *)))) == NULL) {
      (void)fprintf(io->STDERR,
		    "Error: Can't allocate scoring matrix memory.\n\n");
      exit(-1);
    }
    /* Add a row zero only if not TANDEM repeats. */
    if ((p=(int *)calloc(collen * rowlen, sizeof(int))) == NULL) {
      (void)fprintf(io->STDERR,
		    "Error: Can't allocate scoring matrix memory.\n\n");
      exit(-1);
    }

    for (i=0; i<collen; ++i) {
      results->S[i] = p;
      p += rowlen;
    }

    results->S[0][0] = 0;		/* always */
  }
  /*
     Initialize column zero in scoring matrix, for first sequence deletion.
     TANDEM repeat searching uses no column zero.
  */
  if (scoring->btype != TANDEM) {
     if (Sresized) {
        switch (scoring->ptype) {
        case GLOBAL:
	  tscore = scoring->alpha;
	  for (i=1; i<collen; i++) {
	    results->S[i][0] = tscore;
	    tscore += scoring->beta;
          }
          break;
        case LOCAL:
        case FIT:
        case OVER:
             for (i=1; i<collen; i++) {
	        results->S[i][0] = 0;
	     }
          break;
        }
     }
     else if (scoring->memuse == LINEAR) {  /* Re-init S[1][0] */
        switch (scoring->ptype) {
        case GLOBAL:
          if (scoring->stype == PROFILE) {
	     results->S[1][0] = scoring->palpha;
   	     }
          else {
	    results->S[1][0] = scoring->alpha;
	    }
          break;
        case LOCAL:
        case FIT:
        case OVER:
	  results->S[1][0] = 0;
          break;
        }
     }
  }
  /*
     Initialize row zero in scoring matrix, for second sequence deletion.
     If using linear memory, we have to erase leftovers from previous run.
  */
  if (Sresized || (scoring->memuse == LINEAR)) {
    switch (scoring->ptype) {
    case GLOBAL:
    case FIT:
      if (scoring->stype == PROFILE) {
	tscore = scoring->palpha;
	for (j=1; j<rowlen; j++) {
	  results->S[0][j] = tscore;
	  tscore += scoring->pbeta;
	}
      }
      else {
	tscore = scoring->alpha;
	for (j=1; j < rowlen; j++) {
	  results->S[0][j] = tscore;
	  tscore += scoring->beta;
	}
      }
      break;
    case OVER:
    case LOCAL:
      for (j=1; j < rowlen; j++) {
	  results->S[0][j] = 0;
	}
      break;
    }
  }
  /*
     Initialize F vector, for deletions from above.  Since TANDEM scoring
     doesn't use generalized gap(k) = alpha + beta*(k-1), it doesn't use
     E and F matrices.
  */
  if (scoring->btype != TANDEM) {
    if (seq2->resized) {
      if (results->F != (int *)NULL) {
	free(results->F);
      }
      results->F = (int *)calloc(rowlen, sizeof(int));
    }
    infty = gap(seq1->len + seq2->len, scoring);
    if (scoring->mtype == DISTANCE) {
      for (j=0; j<=seq2->len; j++) {
	results->F[j] = infty;  /* We'll be choosing a minimum score */
      }
    }
    else {
      for (j=0; j<=seq2->len; j++) {
	results->F[j] = -infty;  /* We'll be choosing a maximum score */
      }
    }
  }
  /*
      If getting more than one score (declumping) from a matrix...
  */
  /* anticipate repeated alignments, or set for diag in self-repeat */
  if (scoring->naligns > 1 || scoring->btype == SELF) {
    /*
       Initialize matrix of matched/mismatched alignments.
    */
    if (scoring->btype == TANDEM) {
      if (Sresized) {
	if (results->linbits != (char *)NULL) {
	  free(results->linbits);
	}
	if ((results->linbits = calloc(collen,sizeof(char))) == NULL) {
          (void)fprintf(io->STDERR,
			"Error: can't allocate declumping matrix memory.\n\n");
	  exit(-1);
	}
      }
      else {			/* re-use linbits array */
	for (i=0; i<collen; i++)
	  results->linbits[i] = 0;
      }
    }
    else if (scoring->memuse != LINEAR) { /* if non-LINEAR space */
      if (Sresized) {
	if (results->aligned != (char **)NULL) {
	  if (results->aligned[0] != (char *)NULL) {
	    free(results->aligned[0]);
	  }
	  free(results->aligned);
	}
	if ((results->aligned = (char **)malloc(collen*sizeof(char *)))
	    == (char **)NULL) {
	  (void)fprintf(io->STDERR,
			"Error: can't allocate declumping matrix memory.\n\n");
	  exit(-1);
	}
	if ((dtmp = calloc(collen * rowlen, sizeof(char)))
	    == (char *)NULL) {
	  (void)fprintf(io->STDERR,
			"Error: can't allocate declumping matrix memory.\n\n");
	  exit(-1);
	}
	for (i=0; i<collen; i++) {
	  results->aligned[i] = dtmp;
	  dtmp += rowlen;
	}
      }
      else {			/* zero all entries if re-using matrix */
	for (i=0; i<collen; i++)
	  for (j=0; j<rowlen; j++)
	    results->aligned[i][i] = 0;
      }
      if (scoring->btype == SELF) { /* self-repeats; mark off the diagonal */
	for (i=0; i<min(collen,rowlen); i++) {
          results->aligned[i][i] = 1;
          results->S[i][i] = 0;
	}
      }
    }
  }
  /*
     Initialize matrix of top score in each row, and best score for a row/col.
  */
  if (seq1->resized) {
    if (results->best != (int *)NULL) {
      free(results->best);
    }
    results->best = (int *)calloc(seq1->len + 1, sizeof(int));
  }

  results->bestS  = 0;
  results->bestSi = 0;
  results->bestSj = 0;

  if (scoring->stats == PVALUE) { /* Need SEQALN_PVALUE struct initialized */
     /*
        should already be allocated before parseargs.
     */
     if ((unsigned)results->pvalue == (unsigned)NULL) {  /* allocate struct */
        results->pvalue =
            (struct SEQALN_PVALUE *)calloc(1,sizeof(struct SEQALN_PVALUE));
        }
     results->pvalue->pvalue = 0.0;
     results->pvalue->gamma  = 0.0;
     results->pvalue->p      = 0.0;
     results->pvalue->lambda = 0.0;
     results->pvalue->coeff  = 0.0;
     topscore(seq1,seq2,scoring,results,io); /* get maximum score in xmatrix */
     if (results->pvalue->nhist !=
	 results->maxentry * min(seq1->len,seq2->len)) {
       results->pvalue->nhist = results->maxentry * min(seq1->len,seq2->len);
       if (results->pvalue->hist != (int *)NULL) {
	 free(results->pvalue->hist);
       }
       results->pvalue->hist =
         (int *)calloc(results->pvalue->nhist + 1,sizeof(int));
     }
     results->pvalue->npvals = 0;
     if (scoring->nsims >= 0) {
        results->pvalue->nsims = scoring->nsims;
      }
     else {
        results->pvalue->nsims = 0;
      }
     if (scoring->ndeclumps >= 0) {
        results->pvalue->ndeclumps = scoring->ndeclumps;
      }
     else {
        results->pvalue->ndeclumps = 0;
      }
   }
  /* Reset these parameters in case there isn't an intervening load_seq() */
  seq1->resized = 0;
  seq2->resized = 0;

  return;
}
