/* trtrace.c -- construct a traceback list with a tandem repeat matrix */
/*
   Author: Paul Hardy, U.S.C., December 1995.

   Version 1.0 release date: December 1995.
*/

#include <stdio.h>
#include "seqaln.h"

/* Macros for use in the wtrace() subroutine */

#define append_score(X,Y)  (scorel = \
			    (SEQALN_TRACE *)newmem(1, sizeof(SEQALN_TRACE)),  \
                            scorel->i = i,                                    \
			    scorel->j = j,                                    \
                            scorel->next = results->trace,                    \
			    results->trace = scorel)

/*
   Check the match/mismatch (diagonal) direction.
*/
#define scandiag()  if ((results->S[i][j] ==                                  \
			 results->S[i-1][j == 0 ? jmax : j-1] +               \
                         scoring->deltafn(seq1->code[i],                      \
                                   seq2->code[j+1],scoring)) &&               \
		       (!results->linbits || !results->linbits[i])) {         \
		       results->align_len++; i--;			      \
		       j = j==0 ? jmax : j-1;				      \
		       foundalign=1; }

/*
   Scan up from current position.
*/
#define scanup()    if (results->S[i][j] ==				      \
			results->S[i-1][j] + scoring->alpha) {		      \
		       results->align_len++; i--;			      \
		       foundalign = 1; 	/* found a deletion */                \
		    }

/*
   Scan left from current position.
*/
#define scanleft()  if (j==0 && results->S[i][j] ==			      \
        		results->S[i][jmax] + scoring->alpha) {	/* wrap */    \
		       j = jmax;					      \
                       results->align_len++;                                  \
		       foundalign = 1;                                        \
		       } 						      \
                     else if (j > 0 && results->S[i][j] ==		      \
			      results->S[i][j-1] + scoring->alpha) {	      \
                       results->align_len++;                                  \
                       j--;						      \
                       foundalign = 1;                                        \
		       }

/*
   Construct a traceback.

   results->bestS, results->bestSi and results->bestSj must be set before
   this routine is called, and of course the scoring matrix, results->S,
   must have been computed.
*/

void trtrace(seq1, seq2, scoring, results, io)
	struct SEQALN_SEQUENCE *seq1,*seq2;
	struct SEQALN_CONSTANTS *scoring;
	struct SEQALN_RESULTS   *results;
	struct SEQALN_IO *io;
{
int i,j;                /* Position variables for seq1 & seq2, resp. */
int alpha_eq_beta;	/* If scoring->alpha == scoring->beta        */
int Slast, tmplast;	/* For checking multiple character deletions */
int nexti, nextj;  	/* For checking multiple character deletions */
int foundalign;		/* For checking multiple character deletions */
SEQALN_TRACE *scorel;	/* Holding variable for next element in list */
int *current;           /* Points to current position in results->S  */
int leftlen, diaglen, uplen;  /* For adjusting current pointer       */
int jmax;

results->trace = (SEQALN_TRACE *)NULL;
results->align_len = 0;
i = results->bestSi;
j = results->bestSj;
leftlen = 1;
uplen   = seq2->len;
jmax = uplen - 1;
diaglen = uplen + leftlen;

/*
   Set alpha_eq_beta in a separate variable for efficient evaluation.
   Right now, alpha == beta is the only case we handle; we don't maintain
   separate E and F matrices.
*/
alpha_eq_beta = 1;
current = &results->S[i][j];

/*
   Loop until we get a score of 0 -- beginning of an alignment.
   This loop assumes scoring->ptype == LOCAL.
*/
while (results->S[i][j] > 0 && i > 0) {

  if (scoring->Debug) {
    (void)fprintf(io->STDOUT,"(%d,%d)[%d]==>\n",i,j,results->S[i][j]);
  }
  append_score(i,j);
  foundalign = 0;	/* Haven't aligned this letter pair yet */

  /*
    First scan direction.
  */
  if (scoring->envelope) {	/* tracing upper/lower alignment envelope */
    if (scoring->traceupper) {
      scanup();
    }
    else {
      scanleft();
    }
  }
  else {
    scandiag();			/* checking match/mismatch first */
  }

  /*
    Second scan direction.
  */
  if (!foundalign) {
    if (scoring->envelope) {	/* checking match/mismatch second */
      scandiag();
    }
    else {
      if (scoring->traceupper) {
	scanup();
      }
      else {
	scanleft();
      }
    }

    /*
      Third scan direction -- the indel direction we don't prefer.
    */
    if (!foundalign) {
      if (scoring->traceupper) {
	scanleft();
      }
      else {
	scanup();
      }
    }				/* end scan in third direction  */
  }				/* end scan in second direction */

  /*
    Should have found something by now -- checked all three directions.
  */
  /* checked all directions & didn't get anywhere--abort */
  if (!foundalign) {
    (void)fprintf(io->STDERR,"Traceback routine failed at %d %d.\n",i,j);
    (void)fprintf(io->STDERR,
          "   score is %d; should be max(%d+%d,%d %d,%d %d).\n",
                        results->S[i][j],
                        j==0 ? results->S[i-1][jmax] : results->S[i-1][j-1],
                        scoring->deltafn(seq1->code[i], seq2->code[j], scoring),
                        results->S[i-1][j], scoring->alpha,
                        j==0 ? results->S[i-1][jmax] : results->S[i][j-1],
                        scoring->alpha);
    (void)fprintf(io->STDERR, " left = %u;  up = %u;  diag = %u.\n",
                    current - leftlen,  current - uplen,  current - diaglen);
    (void)fprintf(io->STDERR,"Began traceback with score of %d at %d %d.\n",
	    results->bestS, results->bestSi, results->bestSj);
    (void)fprintf(io->STDERR,"Current position is %u; array starts at %u.\n",
            current, results->S[0]);
    print_trace(scoring,results,io);
    i = j = 0;		/* so loop will end, even if LOCAL since S[0][0]==0 */
    results->trace = (SEQALN_TRACE *)NULL;
  }
}	    /* end while */

if (scoring->Debug) {
  (void)fprintf(io->STDOUT,"(%d,%d).\n",i,j);
}

/*
   We encountered impossible values for i and/or j.
*/
if (i < 0 || j < 0) {
  (void)fprintf(io->STDERR,
        "Traceback parameters out of range at %d %d.\n",i,j);
  (void)fprintf(io->STDERR, "Began traceback at %d %d:\n",
	  results->bestSi, results->bestSj);
  print_trace(scoring,results,io);
  results->trace = (SEQALN_TRACE *)NULL;
}

return;
}
